'''tmplr's frontend orchestration'''

import argparse
import os.path as path
import sys

import tmplr
import tmplr.temple
import tmplr_cli
import tmplr_cli.editor as editor


def parser():
    p = argparse.ArgumentParser(
            description='Holy template renderer',
            )
    p.add_argument(
            '--version',
            action='version',
            version=tmplr_cli.__version__,
            )
    p.add_argument(
            '-d',
            '--dir',
            default=path.expanduser(path.join('~', '.tmplr')),
            help='Directory of temples: default ~/.tmplr',
            )
    p.add_argument(
            '--no-edit',
            action='store_true',
            help='Do not edit the rendered temple--just generate it',
            )
    p.add_argument(
            '--stdout',
            action='store_true',
            help='Write rendered temple to stdout; implies --no-edit',
            )
    p.add_argument(
            '-f',
            '--file',
            help='''Provide a filename for {fname}. If no filename is given for
                    {fname}, implies --stdout''',
            )
    p.add_argument(
            '-F',
            '--print-file',
            help='''Do not render temple. Print filename for rendered
                    template (including -f values). Not compatible with any
                    other render options.''',
            action='store_true',
            )
    p.add_argument(
            'temple',
            help='Temple to render',
            # type=tmplr.temple.from_file,
            # can't do this because we depend on -d dir
            )
    p.add_argument(
            'render_args',
            nargs='*',
            help='var=value substitutions to be rendered',
            )
    return p


def parse_kv(kv):
    key, *vals = kv.split('=')
    val = '='.join(vals)
    return (key, val)


def main():
    p = parser()
    args = p.parse_args()
    if args.stdout:
        args.no_edit = True
    temples = tmplr.temple.temples(args.dir)
    if args.temple not in temples:
        p.error('''No temple "{temple}".

Create it with "temples -e -t {temple}".'''.format(temple=args.temple))
    temple = tmplr.temple.from_file(path.join(args.dir, args.temple))
    if args.print_file:
        print(temple.filename(args.file))
        sys.exit(0)
    subs = dict(
            map(
                lambda kv: parse_kv(kv),
                args.render_args))
    if not all(
            key in subs
            for key in temple.placeholders()):
        p.error('''Missing placeholder value.

Run "temples -t {temple}" to see available keys.'''.format(temple=args.temple))
    temple.render(subs)
    written = temple.write(filename=args.file, stdout=args.stdout)
    if written:
        if args.no_edit:
            # log file for application usage
            print(written)
            exit = 0
        else:
            exit = editor.edit(written)
    else:
        exit = 0
    sys.exit(exit)


if __name__ == '__main__':
    main()



'''The holiest cli template system

tmplr provides modules for working with tmplr templates, called Temples. Most
Temples are represented by actual files on disk, so a convenience function is
provided to create a Temple from a path. Programmatic access is also provided.

Modules:
    temple -- provides the Temple class and helper functions

Temple-file format
------------------

A Temple file consists of two parts:
    1. header
    2. content

The header provides metadata specific to tmplr, and will be omitted when the
template is rendered into output.

The content is text, containing special template sequences which will be
substituted by render arguments upon output.

Header format
=============

The first line of the header must consist of the same character repeated
thrice. We call this the "comment character" because, if you use a character
considered a comment for the normal file content, editors will not choke on
Temple headers.

The next lines consist of metadata and all follow this format:
    {comment character} key : value

A value ends once the line is terminated by a newline character.

After the metadata is the last line of the header--it is a carbon copy of the
first line of the header.

Anything after the second sequence of 3 comment characters is considered
content and read literally.

Temple metadata
===============

The supported key-value pairs are:
    - output
        The output directive can be a path (with ~ expansions), optionally
        containing the sequence {fname}. If the rendered template is written,
        it will be written to this path, with {fname} substituted for a
        filename by the engine (see tmplr.temple.Temple.write).

        If output is instead the string 'stdout', the rendered template will be
        printed to standard out.
    - help
        The help directive provides a short description of the template
    - delim
        The delim directive decides the special sequences that require render
        arguments.

        Any text of delim followed by name, or delim{name}, will be replaced by
        the render argument name (see tmplr.temple.Temple.render).

        Take care to choose a delim value that will not appear in the template
        except for in these escape sequences.

Unsupported values will be ignored, but not cause a parsing error.

*** Example ***

###
# output : /tmp/tmplr-test-example-{fname}
# help : tmux config script
# delim : %%
###
#! /bin/sh
content...
%%var
%%{func_name}
%%var
...etc.

*** End Example ***

In this example, the comment character is '#' because the file is a shell
script. %%var will be replaced in both places (since %% is the delim) by the
var argument to render. Similarly %%{func_name} will be replaced the func_name
argument to render.

If fname is passed as, e.g., templar, the result will be written to
/tmp/tmplr-test-example-templar.
'''

__version__ = '0.1.1'



'''temples' frontend orchestration'''

import argparse
import os
import os.path as path
import sys

import tmplr
import tmplr.temple
import tmplr_cli
import tmplr_cli.editor as editor


def parser():
    p = argparse.ArgumentParser(
            description='Queries temples',
            )
    p.add_argument(
            '--version',
            action='version',
            version=tmplr_cli.__version__,
            )
    p.add_argument(
            '-d',
            '--dir',
            default=path.expanduser(path.join('~', '.tmplr')),
            help='Directory of temples: default ~/.tmplr',
            )
    p.add_argument(
            '-e',
            '--edit',
            action='store_true',
            help='Edit the temple',
            )
    p.add_argument(
            '-t',
            '--temple',
            help='Temple to query',
            # type=tmplr.temple.from_file,
            # can't do this because we depend on -d dir
            )
    return p


def main():
    p = parser()
    args = p.parse_args()
    if args.temple is None:
        temples = tmplr.temple.temples(args.dir)
        print('\n'.join(temples.keys()))
        sys.exit(0)
    elif args.edit:
        if not path.exists(args.dir):
            os.makedirs(args.dir)
        temple = path.join(args.dir, args.temple)
        sys.exit(editor.edit(temple))
    else:
        temples = tmplr.temple.temples(args.dir)
        if args.temple not in temples:
            p.error('''No temple "{temple}".

Create it with "temples -e -t {temple}".'''.format(temple=args.temple))
        temple = tmplr.temple.from_file(path.join(args.dir, args.temple))
        print(temple.helptext())
        sys.exit(0)


if __name__ == '__main__':
    main()



#! /usr/bin/env python

from setuptools import setup
import io

import tmplr


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


setup(
    name='tmplr',
    version=tmplr.__version__,
    url='https://github.com/benknoble/tmplr',
    author='D. Ben Knoble',
    author_email='ben.knoble@gmail.com',
    license='MIT',
    description='The holiest cli template system',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Software Development :: Code Generators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        ],
    install_requires=[
        ],
    python_requires='>=3',
    packages=[
        'tmplr',
        'tmplr_cli',
        ],
    test_suite='tests.suite',
    entry_points={
        'console_scripts': [
            'tmplr=tmplr_cli.tmplr_app:main',
            'temples=tmplr_cli.temples_app:main',
            ],
    },
)



'''Classes and functions for dealing with tmplr templates

Classes:
    Temple -- a template

Functions:
    make_Template -- generates a subclass of string.Template on the fly
    from_file -- Temple from file
    temples -- maps names to Temples in a directory
'''

import string
import re
import os
import os.path as path
import glob


def make_Template(name, **kwargs):
    '''make_Template(name, **kwargs) -> subclass of string.Template

    - name is the name of the subclass
    - kwargs is passed as the dict element to type
    '''
    return type(
            name,
            (string.Template,),
            kwargs)


class Temple(object):
    '''Template object for templr

    Temple(name, output, helptext, content, delim='%%') -> new Temple

    Methods:
        helptext() -- text for help on the Temple
        placeholders() -- a list of names needing values for render
        render(subs=None, **kwargs) -- render the template
                                  (see string.Template.safe_substitute)
        write(filename=None) -- write via output
    '''

    def __init__(self, name, output, helptext, content, delim='%%'):
        self.name = name
        self.output = output
        self.help = helptext
        self.content = content
        self.delim = delim
        self.placeholder_re = re.compile(re.escape(delim) + r'{?\w+}?')
        self.Template = make_Template(name, delimiter=delim)
        self.template = self.Template(content)

    def helptext(self):
        '''helptext() -> str

        Text suitable for display as help about the Temple
        '''
        return '%s : %s\nPlaceholders\n\t%s' % (
                self.name,
                self.help,
                '\n\t'.join(self.placeholders()))

    def placeholders(self):
        '''placeholders() -> list

        List of names needing values for render
        '''
        return _orderedset(
                map(
                    lambda p: p.lstrip(self.delim + '{').rstrip('}'),
                    self.placeholder_re.findall(self.template.template)))

    def filename(self, filename):
        '''filename(filename) -> str

        Filename where rendered template *would* be written
        '''
        return path.expanduser(self.output.format(fname=filename))

    def render(self, subs=None, **kwargs):
        '''render(subs=None, **kwargs) -> str

        see string.Template.safe_substitute for the parameters
        '''
        self.rendered = self.template.safe_substitute(subs, **kwargs)
        return self.rendered

    def write(self, filename=None, stdout=False):
        '''write(filename=None) -> path

        Write via output.
            1. If any of
                - stdout is True
                - output is stdout
                - filename is None but {fname} is in output
                then print and return None
            2. Else, write to file(output) & return path written
                - {fname} will be substituted for filename
                - if path already exists, don't write, but return path
        '''
        stdout_conditions = (
                stdout,
                self.output is 'stdout',
                filename is None and '{fname}' in self.output,
                )
        if any(stdout_conditions):
            print(self.rendered, end='')
            return None
        else:
            fullpath = self.filename(filename)
            dirname = path.dirname(fullpath)
            if path.exists(fullpath):
                # silent
                return fullpath
            if not path.isdir(dirname):
                os.makedirs(dirname)
            with open(fullpath, mode='w', errors='strict') as f:
                f.write(self.rendered)
            return fullpath


def from_file(filename):
    '''from_file(filename) -> Temple

    Read filename to produce a Temple. see package documentation for details on
    the format.

    Take care--almost no error handling is done. Bad input will fail noisly.
    '''
    comment_char = ''
    directives = dict()
    content = ''
    with open(filename, mode='r', errors='strict') as f:
        firstline = f.readline()
        comment_char = firstline[0]
        for line in f:
            if line == firstline:
                break
            (directive, value) = map(
                    lambda s: s.strip(),
                    line.lstrip(comment_char).split(':'))
            directives[directive] = value
        content = f.read()
    return Temple(
            path.basename(filename),
            directives['output'],
            directives['help'],
            content,
            delim=directives['delim'])


def temples(tpath):
    '''temples(tpath) -> dict

    The result is a mapping of template names to Temple values.  If tpath is a
    directory, this mapping contains a Temple for each file in the directory.
    If it is a single file, it only contains that file.

    See also from_file.
    '''
    if not path.exists(tpath):
        return dict()
    elif path.isdir(tpath):
        files = glob.iglob(
                path.join(
                    path.expanduser(tpath),
                    '*'))
    elif path.isfile(tpath):
        files = [tpath]
    else:
        return dict()
    return {t.name: t
            for t in map(
                from_file,
                files)}


def _orderedset(keys):
    return list(dict.fromkeys(keys).keys())



'''Provides less-like edit functions for invoking the user's editor'''

import os
import subprocess


def edit(filename):
    '''edit(filename) -> return code

    Edit a file. cf. editor in this module
    '''
    cmd = [editor(), filename]
    return subprocess.run(cmd).returncode


def editor():
    '''editor() -> str

    In the same way as ``less`` and many other programs, it tries the
    environment variables VISUAL and EDITOR (in that order) before defaulting
    to vi.
    '''
    defaults = ['vi']
    env_vars = [
            'VISUAL',
            'EDITOR'
            ]
    editors = [
            ed
            for ed in map(
                lambda e: os.environ.get(e),
                env_vars)
            if ed is not None
            ] + defaults
    assert len(editors) > 0, "No editors found"
    return editors[0]
